# version 2022.1
[CmdletBinding()]
param (
	[string]$logFileName,
	[switch]$logToConsole
)

$Global:ErrorActionPreference = 'Stop'
$Global:mmConfig = [mm]::parseConfigFile((Join-path $PSScriptRoot 'settings.cfg'))
[MMAPI]$Global:mmAPIClient = $null # [mm]::getAPI sets this

$verbosity = if ($VerbosePreference -eq 'Continue') { 3 } else { $Global:mmConfig.logging.verbosity }
if ($logFileName){
	$Global:mmLogger = [Logger]::new($logFileName, $logToConsole, $verbosity)
}
$Global:mmCache = [Cache]::new()


function isEmpty([string] $string){
	return [mm]::isEmpty($string)
}

function notEmpty ([String]$string){
	return [mm]::notEmpty($string)
}

function Coalesce($a, $b) { if ([mm]::isEmpty($a)) { $b } else { $a } }
New-Alias "??" Coalesce -Force


function log([string]$msg){
	if ($Global:mmLogger){
		$Global:mmLogger.log($msg)
	} else {
		Write-Host $msg
	}
}

function logv2([string]$msg){
	if ($Global:mmLogger){
		$Global:mmLogger.logv2($msg)
	} elseif ($Global:mmConfig.logging.verbosity -gt 1) {
		write-host "$msg" -ForegroundColor Yellow -BackgroundColor Black
	}
}

function logv3([string]$msg){
	if ($Global:mmLogger){
		$Global:mmLogger.logv3($msg)
	} elseif ($Global:mmConfig.logging.verbosity -gt 2) {
		write-host "$msg" -ForegroundColor Cyan -BackgroundColor Black
	}
}

function logException([string]$msg){
	if ($Global:mmLogger){
		$Global:mmLogger.logException($msg)
	} else {
		Write-Host -ForegroundColor 'Red' -BackgroundColor Black ("Got exception while {0}: {1}" -f $msg, $_.Exception.GetBaseException().Message)
	}
}

function tic ($action){
	if ($Global:mmLogger){
		$Global:mmLogger.tic($action)
	} else {
		log $action
	}
}

function toc ($customAction=$null){
	if ($Global:mmLogger){
		$Global:mmLogger.toc($customAction)
	} elseif ($customAction) {
		log $customAction
	}
}



function returnOutputToCentral ([hashtable]$inNewPropertiesMap = $null, [string]$errorMsg, [int]$errorCode = 1){
	function mmGetCentralXML (){
		$outxml = New-Object "System.Xml.XMLDocument"
		$decl = $outxml.CreateXmlDeclaration("1.0", $null, $null)
		$outxml.InsertBefore($decl, $outxml.DocumentElement) > $null
		$resultEl = $outxml.CreateElement('result')
		if ([mm]::notEmpty($errorMsg)){
			$resultEl.setAttribute('success','0') > $null
			$errorEl = $outxml.createElement('error')
			$errorEl.setAttribute('code',$errorCode.toString()) > $null
			$errorEl.setAttribute('message',$errorMsg) > $null
			$resultEl.appendChild($errorEl) > $null
		} else {
			$resultEl.setAttribute('success','1') > $null
			$newcfs =  $resultEl.OwnerDocument.ImportNode($cfs,$true)
			if ($null -ne $inNewPropertiesMap -and $inNewPropertiesMap.Count -gt 0 ){
				foreach ($propName in $inNewPropertiesMap.Keys){
					try {
						$cNode = $newcfs.SelectSingleNode(("customField[@customFieldName='{0}']" -f $propName))
						$cNode.setAttribute('value',$inNewPropertiesMap[$propName]) > $null
					} catch {
						continue
					}
				}
			}
			$resultEl.appendChild($newcfs) > $null
		}
		$outxml.appendChild($resultEl) > $null
		return [string]$outxml.outerxml
	}


	try {
		try {
			[console]::BufferWidth=10000
		} catch {}

		$xmlstr = mmGetCentralXML
		[Console]::Out.Write($xmlstr)
	} catch {
		logException "while returning error xml"
	}
}

function returnErrorToCentral ([string]$errorMsg, [int]$errorCode=1) {
	log "Exiting and returning error `"$errorMsg`" and code $errorCode to Central"
	returnOutputToCentral -errorMsg $errorMsg -errorCode $errorCode
	[mm]::logoutAndExit()
}

function returnSuccessToCentral ([hashtable]$newProperties){
	try {
		if ($newProperties.Count -gt 0){
			log ('Writing new properties to Central: {0}' -f [mmJSON]::stringify($newProperties))
			returnOutputToCentral -inNewPropertiesMap $newProperties
		}
	} catch{
		logException "while returning success xml"
	}
}


function groupBy ($property) {
	Begin {
		$h=@{};
	}

	Process {
		$key = $_[$property]
		if ([mm]::notEmpty($key)){
			if (-not $h.ContainsKey($key)){
				$h[$key]= New-Object System.Collections.ArrayList
			}
			[void]($h[$key].Add($_))
		}
	}

	End {
		return $h
	}
}

function mapBy ($property) {
	Begin {
		$h=@{};
	}

	Process {
		$key = if ($property) { $_[$property] } else { $_ }
		if ([mm]::notEmpty($key)){
			if (-not $h.ContainsKey($key)){
				$h[$key] += $_
			}
		}
	}

	End {
		return $h
	}
}


class mmJSON {
	static [hashtable]parse([string]$jsonStr) {
		if ($Global:PSVersionTable.PSVersion.Major -lt 6){
			if (-not $Global:SystemWebExtensionsAssemblyLoaded){
				[void]([System.Reflection.Assembly]::LoadWithPartialName("System.Web.Extensions"))
				$Global:SystemWebExtensionsAssemblyLoaded = $true
			}

			$ser = New-Object System.Web.Script.Serialization.JavaScriptSerializer
			$ser.MaxJsonLength = [System.Int32]::MaxValue
			return $ser.Deserialize($jsonStr, [hashtable])
		} else {
			# only PS version 6 has -asHashTable param, as well as fixed maxJsonLength
			return (ConvertFrom-Json -AsHashtable -InputObject $jsonStr)
		}
	}
	static [string]stringify([hashtable]$obj){
		if ($Global:PSVersionTable.PSVersion.Major -lt 6) {
			if (-not $Global:SystemWebExtensionsAssemblyLoaded){
				[void]([System.Reflection.Assembly]::LoadWithPartialName("System.Web.Extensions"))
				$Global:SystemWebExtensionsAssemblyLoaded = $true
			}
			$ser = New-Object System.Web.Script.Serialization.JavaScriptSerializer
			$ser.MaxJsonLength = [System.Int32]::MaxValue
			try {
				return $ser.Serialize(([hashtable]$obj))
			} catch {
				throw ("Could not convert to JSON: {0}" -f $_.Exception.Message)
			}

		} else {
			return ConvertTo-Json $obj -Depth 100
		}
	}
}

class mmException : System.Net.WebException {
	mmException([string]$msg) : base($msg) {}
}

class MMAPI {
	hidden [PSCredential]$credentials
	hidden [string]$session

	[string]$mmWS = 'http://localhost/mmws' # default
	[string]$server = '' # uses whatever mmws has as a default, by default
	[int]$timeout = 120000
	[boolean]$sso
	[string]$username

	MMAPI([string]$username, [string]$password, [string]$mmWS){
		$this.setCredential($username, $password)
		$this.setMMWS($mmWS)
	}

	MMAPI([string]$username, [string]$password){
		$this.setCredential($username, $password)
	}

	MMAPI([string]$mmWS){
		$this.setCredential($null, $null)
		$this.setMMWS($mmWS)
	}

	setCentralServer([string]$centralServer){
		if ($centralServer -and $centralServer.trim()){
			$this.server = $centralServer.trim()
		}
	}

	setTimeout([int]$seconds){
		$this.timeout = $seconds*1000
	}

	hidden setMMWS([string]$mmWS){
		if ($mmWS -and $mmWS.trim()){
			$this.mmWS = $mmWS.trim()
		}
	}

	hidden setCredential([string]$username, [string]$password){
		if ($username -and $password){
			$this.credentials = New-Object System.Management.Automation.PSCredential -ArgumentList $username, (ConvertTo-SecureString -AsPlainText -Force -String $password)
			$this.username = $username
			$this.sso = $false
		} else {
			$this.credentials = $null
			$this.sso = $true
		}
	}

	[boolean] isLoggedIn(){
		return [mm]::notEmpty($this.session)
	}

	[string]login(){
		if ($this.sso){
			tic "Connecting to M&M API with SingleSignOn"
			$params = @{
				server = $this.server
				singleSignOn = $true
			}
		} else {
			tic ("Connecting to M&M API as {0}..." -f $this.credentials.UserName)
			$params = @{
				loginName = $this.credentials.UserName
				password  = $this.credentials.GetNetworkCredential().Password
				server    = $this.server
				unauthorizedAsForbidden = $true
			}
		}
		$resp = $this.call('Login', $params)
		toc "Connected to API"

		$this.session = $resp.session
		if ($resp.userName){
			# sso returns the user name of logged in user
			$this.username = $resp.userName
		}
		return $resp.session
	}

	logout() {
		if ($this.isLoggedIn()){
			log ('Logging out of M&M')
			try {
				$this.call('Logout')
			} catch {
				logException 'while logging out'
			} finally {
				$this.session = ''
			}
		}
	}

	hidden [string]sendRequest([string]$uri, [string]$body){
		$req = [System.Net.HttpWebRequest]::CreateHttp($uri)
		$req.Timeout 	= $this.timeout
		$req.ContentType= "application/json;charset=`"ISO-8859-1`""
		$req.Accept		= "application/json"
		$req.Method		= "POST"
		$req.Proxy		= $null
		$req.Credentials = if ($this.sso) { $null } else { $this.credentials }
		$req.UseDefaultCredentials = $this.sso
		$req.PreAuthenticate = $this.sso

		logv3 ($uri + ':' + $body)

		$buffer = [Text.Encoding]::GetEncoding('iso-8859-1').GetBytes($body)
		$reqStream = $req.GetRequestStream()
		$reqStream.Write($buffer,0,$buffer.length)
		$reqStream.Flush()
		$reqStream.Close()

		try {
			$resp = $req.GetResponse()
		} catch [System.Net.WebException]{
			#because mmws returns a HTTP error on errors, but still includes the error details in the body, we need to fetch the error from the error variable
			$resp = $_.Exception.GetBaseException().Response
			if (!$resp){
				# throw if there's no Response for some reason, so we don't mask the real exception that's causing it
				throw
			}
		}

		$respStream = $resp.GetResponseStream()
		$respReader = [System.IO.StreamReader]$respStream
		$respString = [String]($respReader.ReadToEnd())
		$respStream.Close()
		return $respString
	}

	[hashtable]call([string]$command,[hashtable]$params){
		$mmWSUri = $this.mmWS
		$uri = "$mmWSUri/api/command/$command"

		if ($command -notin @('Login','session')){
			$params.session = if ($this.session) { $this.session } else { $this.login() }
		}

		$responseStr = $this.sendRequest($uri, [mmJSON]::stringify($params))
		if ([mm]::isEmpty($responseStr)){
			logv3 "<blank response from API>"
			return @{}
		}

		logv3 $responseStr
		try {
			$response = [mmJSON]::parse($responseStr)
		} catch {
			$response = $null
		}
		if ($response.error){
			throw [mmException]::new($response.error.message)
		}
		if ($response.result){
			return $response.result
		}
		throw [System.Net.WebException]::new($responseStr)
	}
	[hashtable]call([string]$command){
		return $this.call($command, @{})
	}

	[void]callAsync([string]$command,[hashtable]$params){
        if (!$this.isLoggedIn()){
            $this.login()
        }
		$prevTimeout = $this.timeout
		$this.timeout = 50
		try {
			$this.call($command,$params)
		} catch {} # all errors ignored, including timeout, and no return value collected
		$this.timeout = $prevTimeout
	}


	[hashtable[]]executePreparedStatement([string]$inStatement,[hashtable]$inParamsMap){
		function parseSQLResult($sqlResult){
			$rows = New-Object System.Collections.ArrayList
			foreach ($row in $sqlResult.rows){
				$i=0;
				$outRow = @{}
				foreach ($header in $sqlResult.header){
					$outRow[$header] = $row[$i]
					$i++
				}
				[void]$rows.Add($outRow)
			}
			return @($rows)
		}

		$params = @()
		foreach ($key in $inParamsMap.Keys){
			$params += @{name=$key; value=$inParamsMap[$key]}
		}
		$res = $this.call('executePreparedStatement', @{name=$inStatement; params=$params})

		return (parseSQLResult $res)
	}

	[hashtable]getCustomProperties([string]$ref, [string]$objType,[boolean]$onlyCustom){
		$properties = $this.call('GetProperties', @{ref=$ref; objType=$objType; onlyCustom=$onlyCustom}).properties
		return [mm]::getPropertyMap($properties)
	}
	[hashtable]getCustomProperties([string]$ref, [string]$objType){
		return $this.getCustomProperties($ref, $objType, $true)
	}

	setProperties([string]$ref, [string]$objType, [hashtable]$properties, [string]$saveComment, [boolean]$skipEmpty){
		$params = @{ref = $ref; objType = $objType; properties = [mm]::getAsPropertyArray($properties, $skipEmpty)}
		if ([mm]::notEmpty($saveComment)){
			$params.saveComment = $saveComment
		}
		$this.call('SetProperties', $params)
	}
	setProperties([string]$ref, [string]$objType, [hashtable]$properties, [string]$saveComment){
		$this.setProperties($ref, $objType, $properties, $saveComment, $true)
	}
	setProperties([string]$ref, [string]$objType, [hashtable]$properties){
		$this.setProperties($ref, $objType, $properties, $null)
	}
}

class mm {

	static [Boolean]isEmpty($obj){
		if ($obj -is [string]) {
			return $obj.Trim().Length -eq 0
		}
		return $obj.Count -eq 0
	}

	static [Boolean]notEmpty($obj){
		return !([mm]::isEmpty($obj))
	}

	static [string]getCurrentDate () {
		return (get-date -Format 'yyyy-MM-dd HH:mm:ss')
	}
	static [string]getCurrentFilenameDate () {
		return (get-date -Format 'yyyy-MM-dd_HH_mm_ss')
	}

	static [string]getDateAsString([DateTime]$date){
		return (get-date $date -Format 'yyyy-MM-dd')
	}

	static [string]getCurrentDay() {
		return (get-date -Format 'yyyy-MM-dd')
	}

	static [string]getCurrentTime () {
		return (get-date -Format 'HH:mm:ss')
	}

	static [string]getTimestamp (){
		return ((get-date -UFormat %s) -split '[.,]')[0]
	}

	static [hashtable]getPropertyMap($inProperties){
		if ($inProperties.property -is [array]) {
			# the JSON received by external scripts hasn't been cleaned up, so the child elements are actually the properties
			$inProperties = $inProperties.property
		} elseif ($inProperties.property.name) {
			# the same, except it wasn't written to json as array when a single property
			$inProperties = @($inProperties.property)
		}
		if ($inProperties -is [array]){
			$map = @{}
			foreach ($prop in $inProperties){
				$map[$prop.name]=$prop.value
			}
			return $map
		}
		# assume it's already a map then
		return $inProperties
	}

	static [hashtable[]]getAsPropertyArray([hashtable]$propMap, [boolean]$skipEmpty){
		$arr = @()
		foreach ($pIter in $propMap.GetEnumerator()){
			if (!$skipEmpty -or [mm]::notEmpty($pIter.value)){
				$arr += @{name = $pIter.name; value = $pIter.value}
			}
		}
		return $arr
	}

	static [hashtable[]]getAsPropertyArray([hashtable]$propMap){
		return [mm]::getAsPropertyArray($propMap, $false)
	}

	static [string[]]getUniqueValues ([string[]]$inValues){
		$outValues=@{}
		foreach ($item in $inValues){
			if ([mm]::notEmpty($item)){
				$outValues[$item]=$true
			}
		}
		return @($outValues.Keys)
	}

	static [hashtable]removeEmptyHashtableKeys ([hashtable]$inMap){
		$newMap = @{}
		foreach ($key in $inMap.Keys){
			if ([mm]::notEmpty($inMap[$key]) -and $inMap[$key].Count -gt 0){
				$newMap[$key]=$inMap[$key]
			}
		}
		return $newMap
	}

	static [string]ref2ID([string]$inRef){
		if ([mm]::isEmpty($inRef)){
			return $null
		}

		if ($inRef -match '\{#\d+-#(\d+)\}'){
			return $matches[1]
		}
		if ($inRef -match '^[A-Za-z]+/(\d+)$'){
			return $matches[1]
		}
		return $null
	}

	static [string]ID2ObjRef ($inID,$inType){
		return ("{{#{0}-#{1}}}" -f $inType,$inID)
	}

	static [string]clipDomainName ([string]$inString,[string]$inDomainName){
		return ($inString -replace [Regex]::escape($inDomainName),'').trimend('.').toLower()
	}

	static [string]cleanXmlString ([string]$str){
		return $str.replace('&#x5C;','\')
	}

	static [hashtable]parseUserName($inName, [boolean]$requireDomainPrefix){
		$userParts = $null
		if ($inName.split('\').Length -eq 2){
			$p = $inName.split('\')
			$userParts = @{'name'=$p[1];'domain'=$p[0]}
		} elseif ($inName.split('@').Length -eq 2){
			$p = $inName.split('@')
			$userParts = @{'name'=$p[0];'domain'=$p[1]}
		} elseif (!$requireDomainPrefix) {
			logv2 ('Assuming domain for user {0} is {1}' -f $inName, $env:USERDOMAIN)
			$userParts = @{'name'=$inName;'domain'=$env:USERDOMAIN}
		}
		return $userParts
	}

	static [hashtable]parseUserName($inName){
		return [mm]::parseUserName($inName, $false)
	}

	static [string]replaceTokens ([string]$templateString,[hashtable]$tokenMap){
		$oStr = $templateString
		foreach( $token in $tokenMap.GetEnumerator() ){
			$pattern ='{{{0}}}' -f $token.key
			$oStr = $oStr -replace $pattern, $token.Value
		}
		return $oStr
	}

	static [boolean]respondsToPing ([string]$inIP,[int]$timeout){
		$obj = New-Object system.Net.NetworkInformation.Ping
		$result = $obj.Send($inIP, $timeout)
		return $result.status -eq 'Success'
	}
	static [boolean]respondsToPing ([string]$inIP){
		return [mm]::respondsToPing($inIP, 700)
	}

	static [object]resolveName([string]$inName,[string]$inServer){
		$server = ?? $inServer $Global:mmConfig.DNS.resolver
		log ('Resolving {0} with {1}' -f $inName, $server)
		$res = Resolve-DnsName -Name $inName -Server $server -DnsOnly -QuickTimeout -DnssecCd -ErrorAction SilentlyContinue
		$ret = if ($?) { $res } else { $null }
		return $ret
	}

	static [string[]]resolveNameToIP([string]$inName, [string]$inServer){
		$resp = [mm]::resolveName($inName, $inServer)
		if ($resp){
			$IPs = @()
			foreach($item in $resp){
				if ($item.IP4Address){
					$IPs += $item.IP4Address
				}
			}
			foreach($item in $resp){
				if ($item.IP6Address){
					$IPs += $item.IP6Address
				}
			}
			return $IPs
		}
		return $false
	}

	static [boolean]compareRecords([hashtable]$rec1,[hashtable]$rec2){
		return 	$rec1.name -eq $rec2.name -and `
				$rec1.ttl  -eq $rec2.ttl -and `
				$rec1.type -eq $rec2.type -and `
				$rec1.data -eq $rec2.data -and `
				$rec1.dnsZoneRef -eq $rec2.dnsZoneRef
	}

	static [hashtable]getChangesFromRequest([hashtable]$request, [string[]]$inStates) {
		$changesMap = @{}
		foreach ($field in $request.Keys){
			if ($field.endsWith('Changes')){
				$childField = $field.trimEnd('s')
				if ($request.$field.$childField){
					# e.g. request.dnsRecordChanges.dnsRecordChange, is array when more than one, but object when single. 
					# Wrapping with @() makes it always an array (doesn't nest array if already array)
					$changes = @($request.$field.$childField)
				} else {
					$changes = @($request.$field)
				}

				$changesMap.$field = if ($inStates.Count) { $changes | Where-Object state -in $inStates } else { $changes }
			}
		}
		return $changesMap
	}
	static [hashtable]getChangesFromRequest([hashtable]$request){
		return [mm]::getChangesFromRequest($request, @())
	}

	static [hashtable[]]getFlatChangeListFromRequest([hashtable]$request, [string[]]$inStates) {
		[hashtable]$changesMap = [mm]::getChangesFromRequest($request, $inStates)

		$changes = @()
		foreach ($field in $changesMap.Keys) {
			$changes += $changesMap.$field #yes, this concats the array
		}
		return $changes
	}
	static [hashtable[]]getFlatChangeListFromRequest([hashtable]$request){
		return [mm]::getFlatChangeListFromRequest($request, @())
	}

	static [hashtable]parseConfigFile([string]$FilePath){
		$CommentCount = 0
		$ini = @{}
		$section = $null
		foreach ($line in (Get-Content $FilePath)){
			if ($line -match "^\[(.+?)\]"){ # Section
				$section = $matches[1]
				$ini[$section] = @{}
				$CommentCount = 0
			} elseif ($line -match "^(\s*;.*)$"){
				if (!$section){
					$section = "No-Section"
					$ini[$section] = @{}
				}
				$value = $matches[1].trim()
				$CommentCount = $CommentCount + 1
				$name = "Comment" + $CommentCount
				$ini[$section][$name] = $value
			} elseif ($line -match "(.+?)\s*=\s*(.*)$"){ #key
				if (!$section){
					$section = "No-Section"
					$ini[$section] = @{}
				}
				$name,$value = $matches[1..2]

				if ($value.trim() -match '^"(.*?)"' -or $value.trim() -match "^'(.*?)'"){
					$value = $matches[1].trim()
				} else {
					$value = $value.split(';')[0].trim()
				}

				if (@('false','no','0') -contains $value){
					$value = $false
				} elseif (@('true','yes','1') -contains $value){
					$value = $true
				} elseif ($value -eq ''){
					$value = $null
				}

				$ini[$section][$name.trim()] = $value
			}
		}
		return $ini
	}

	static storeDebugBlob([string]$str, [string]$prefix, [string]$extension){
		if ($Global:mmConfig.logging.storeStdIn) {
			mkdir ./tmp/inputs -force -ErrorAction Ignore | out-null
			if ([mm]::isEmpty($prefix)){
				$prefix = 'input'
			}
			$tmpFileName = '.\tmp\inputs\{0}-{1}.{2}' -f $prefix,[mm]::getCurrentFilenameDate(),$extension
			$str | out-file -FilePath $tmpFileName
			logv2 "storing script input in $tmpFileName"
		}
	}

	[MMAPI]static getAPI([string]$username, [string]$password){
		$config = $Global:mmConfig.API
		$client = [MMAPI]::new($username, $password, $config.mmWS)
		$client.setCentralServer($config.CentralServer)
		$Global:mmAPIClient = $client
		return $client
	}

	static logoutAndExit(){
		if ($Global:mmAPIClient){
			$Global:mmAPIClient.logout()
		}
		exit 0
	}
}

class mmEmail {

	static [string]send([string[]]$toAddresses,[string]$subject,[string]$body,[string]$cc, [string]$bcc) {
		### SMPT Client configuration (SSL, Credentials, etc). Change this according to what the SMPT server specified above supports
		function getAddressList([string]$emailsStr){
			if (!$emailsStr){
				return @()
			} else {
				return ($emailsStr -split '[;,]').trim()
			}
		}
		$config = $Global:mmConfig.email
		$bannerImgSrc = ?? $config.bannerImgSrc 'cid:mmlogo.png'
		$client = new-object Net.Mail.SmtpClient $config.smtpServer
		$client.UseDefaultCredentials = ?? $config.UseDefaultCredentials $false
		$client.EnableSsl = ?? $config.EnableSsl $false
		if ($config.username -and $config.password){
			$client.UseDefaultCredentials = $false
			$client.Credentials = New-Object System.Net.NetworkCredential($config.username, $config.password)
		}

		#Creating a Mail object
		$msg = new-object Net.Mail.MailMessage
		$msg.From = $config.fromAddress
		$msg.ReplyTo = ?? $config.ReplyTo $config.fromAddress
		foreach ($bccAddr in (getAddressList $bcc)){
			$msg.Bcc.Add($bccAddr)
		}
		foreach ($ccAddr in (getAddressList $cc)){
			$msg.CC.Add($ccAddr)
		}
		foreach ($email in $toAddresses){
			if ([mm]::notEmpty($email)){
				$msg.To.Add($email)
			}
		}
		$msg.subject = $subject
		$msg.body = [mm]::replaceTokens($body, @{bannerImgSrc = $bannerImgSrc})
		$msg.isBodyHTML= $true
		$msg.BodyEncoding = [System.text.Encoding]::UTF8


		if ($bannerImgSrc.ToLower().StartsWith('cid:')){
			# we got a local file here. We need to attach it
			$bannerFileName = $bannerImgSrc.split(':')[1]
			$attch = New-Object Net.Mail.Attachment((join-path $PSScriptRoot "email-templates/$bannerFileName"))
			$attch.ContentId = $bannerFileName
			$attch.ContentDisposition.Inline = $true
			$attch.ContentDisposition.DispositionType = 'Inline'
			$attch.ContentDisposition.FileName = $bannerFileName
			$attch.ContentType = "image/png"
			$msg.Attachments.Add($attch)
		}


		$retryTimes = ?? $config.retries 5
		$retryInterval = ?? $config.retryInterval 500

		$f = 0
		while ($true){
			try {
				$client.Send($msg)
				break
			} catch {
				#when there's an error, like "too many messages", let's try again after we wait (blindly) for half a second, for max 5 tries
				if ($f++ -lt $retryTimes){
					start-sleep -Milliseconds $retryInterval
				} else {
					throw
				}
			}
		}
		return $msg.body
	}

	static prepareAndSend([string[]]$toAddresses, [string]$subject, [string]$templateFilepath, [string]$cc, [string]$bcc, [hashtable]$tokenMap) {
		cd $PSScriptRoot
		$config = $Global:mmConfig.email
		$relPath = './email-templates'
		$header = (Get-Content "$relPath/email-header.htm" -Raw)
		$footer = (Get-Content "$relPath/email-footer.htm" -Raw)
		$theBody = (Get-Content "$relPath/$templateFilepath" -Raw)

		if ($null -eq $tokenMap){
			$tokenMap = @{}
		}
		$tokenMap.webUri = $config.webUri

		$theBody = [mm]::replaceTokens($theBody, $tokenMap)
		$header = [mm]::replaceTokens($header, @{webUri = $config.webUri})
		$footer = [mm]::replaceTokens($footer, @{webUri = $config.webUri})


		$theAddresses = $toAddresses

		$completeBody = ($header + $theBody + $footer)
		if ([mm]::notEmpty($config.forceTo)){
			$msg = ('Email redirected to {0} for testing purposes (was to:{1}, cc:{2}, bcc:{3})' -f $config.forceTo,($toAddresses -join ','),$cc,$bcc)
			log $msg
			$completeBody = $completeBody.Replace('</body>',"<p style='font-size:10pt'>$msg</p>`n</body>")
			$theAddresses = @($config.forceTo)
			$cc = $null
			$bcc = $null
		}

		$sentBody = $completeBody
		try {
			$sentBody = [mmEmail]::send($theAddresses, $subject, $completeBody, $cc, $bcc)
			log 'Email successfully sent'
			$sendResult= 'Successfully sent'
		} catch {
			log ('Got error while sending email: {0}' -f $_.Exception.Message)
			$sendResult='Error sending:' +$_.Exception.Message
		}

		if ($config.debug) {
			try {
				if (!(Test-Path "$relPath/debug")){
					mkdir "$relPath/debug"
				}
				$logbd = "Subject: {0}`r`nTo: {1} (originally {2})`r`n{3}`r`nTemplate: {4}`r`n`r`nEmail Body:`r`n{5}" -f $subject,($theAddresses -join ','),($toAddresses -join ','),$sendResult,"$relPath/$templateFilepath",$sentBody.replace("`n","`r`n").Replace("`r`r","`r")
				$logbd | out-file ('{0}/debug/{1}-{2}.txt' -f $relPath,$templateFilepath.split('/')[-1],[mm]::getTimestamp())
			} catch {
				logException 'while logging debug email'
			}
		}
	}

	static prepareAndSend ([string[]]$toAddresses, [string]$subject, [string]$templateFilepath, [hashtable]$tokenMap) {
		[mmEmail]::prepareAndSend($toAddresses, $subject, $templateFilepath, $null, $null, $tokenMap)
	}
	static prepareAndSend ([string[]]$toAddresses, [string]$subject, [string]$templateFilepath) {
		[mmEmail]::prepareAndSend($toAddresses, $subject, $templateFilepath, $null, $null, @{})
	}

}

class htmlUtils {
	static [string]removePotentialURLStylingHTML ($inHTMString){
		return ('<span class="appleLinksBlack">{0}</span>' -f $inHTMString)
	}

	static [string]getStateAsHTMLBlob([hashtable]$inChange) {
		$tpl = '<span style="color:{0};">({1})</span>'
		$color = 'black'

		if ($inChange.state -eq 'Failed'){
			$color='red'
		} elseif ($inChange.state -eq 'Applied'){
			$color='green'
		}
		return ($tpl -f $color,$inChange.state)
	}

	static [string]getPropertiesAsListItems([hashtable]$propertyMap){
		$str = ''
		foreach ($name in $propertyMap.Keys){
			$value = $propertyMap[$name]
			$str += "<li><b style=`"margin-right:4px;`">$name</b>: $value</li>"
		}
		return $str
	}

	static [string]getRecordChangeAsHTML([hashtable]$change, [boolean]$includeState){
		$state = if ($includeState) { [htmlUtils]::getStateAsHTMLBlob($change) } else { '' }

		$action = $change.type.toString().replace('DNSRecord','')
		
		# add and delete are straightforward, just print the record being added or removed, along with the properties
		if ($action -eq 'Add'){
			$r = $change.dnsRecordAfter
			$props = [mm]::getPropertyMap($r.customProperties)
			$AddRecordFormat = '[<b style="color:green">{0}</b>]: "{1}" {2}<ul>{3}</ul>'
			return ($AddRecordFormat -f 'Add',[htmlUtils]::getRecordAsHTML($r),$state,[htmlUtils]::getPropertiesAsListItems($props))
		}
		if ($action -eq 'Remove') {
			$r = $change.dnsRecordBefore
			$props = [mm]::getPropertyMap($r.customProperties)
			$DelRecordFormat = '[<b style="color:red">{0}</b>]: "{1}" {2}<ul>{3}</ul>'
			return ($DelRecordFormat -f 'Delete',[htmlUtils]::getRecordAsHTML($r),$state,[htmlUtils]::getPropertiesAsListItems($props))
		}

		# modify is a bit more tricky, in particular with the added ability of being able to change record custom properties as part of the change 
		# the record itself may not be changing, but the properties may also be changing, and we need to include that in the description
		# also, we don't want to state something -> something when it doesn't change
		# the same for the properties!
		$b = $change.dnsRecordBefore
		$a = $change.dnsRecordAfter
		$recStr = ''  
		if ([mm]::compareRecords($a,$b)){
			# record did not change, only it's custom properties (presumably)
			$recStr = '"{0}"' -f [htmlUtils]::getRecordAsHTML($a)
		} else {
			$recStr = '"{0}" to "{1}"' -f [htmlUtils]::getRecordAsHTML($b,$true),[htmlUtils]::getRecordAsHTML($a,$true)
		}

		$propStr = ''
		$bProps = [mm]::getPropertyMap($b.customProperties)
		$aProps = [mm]::getPropertyMap($a.customProperties)
		foreach ($propName in $bProps.Keys){
			$bVal = $bProps[$propName]
			$aVal = $aProps[$propName]

			$propStr += "<li><b style=`"margin-right:4px;`">$propName`:</b>"
			if ($bVal -eq $aVal) {
				$propStr +=  "$bVal</li>"
			} else {
				$propStr += "$bVal &rarr; $aVal </li>"
			}
		}

		$ModifyRecordFormat = '[<b>{0}</b>]: {1} {2}<ul>{3}</ul>'
		return ($ModifyRecordFormat -f 'Modify',$recStr,$state,$propStr)
	}

	static [string]getRecordChangeListAsHTML ([hashtable[]]$changes,[boolean]$includeState) {
		$items = $changes | ForEach-Object { return [htmlUtils]::getRecordChangeAsHTML($_, $includeState) }
		return [htmlUtils]::getArrayAsHTMListItems($items)
	}

	static [string]getRecordChangeListAsHTML ([hashtable[]]$changes){
		return [htmlUtils]::getRecordChangeListAsHTML($changes, $false)
	}

	static [string]getRecordAsHTML([hashtable]$inRecord,[boolean]$includeTTL){
		$s = '&nbsp;&nbsp;&nbsp;'
		$fqdn = if ([mm]::notEmpty($inRecord.fqdn)) { $inRecord.fqdn } else { $inRecord.name }
		$ttlBlob = if ($includeTTL) { "$($inRecord.ttl)$s"} else { "" }
		$str = "{0}$s{1}{2}$s{3}" -f $fqdn,$ttlBlob,$inRecord.type,$inRecord.data
		return [htmlUtils]::removePotentialURLStylingHTML($str)
	}
	static [string]getRecordAsHTML([hashtable]$inRecord){
		return [htmlUtils]::getRecordAsHTML($inRecord, $false)
	}

	static [string]getRecordListAsHTMLListItems ([hashtable[]]$records){
		$items = $records | ForEach-Object {return [htmlUtils]::getRecordAsHTML($_, $true)}
		return [htmlUtils]::getArrayAsHTMListItems($items)
	}

	static [string]getRecordAsString ([hashtable]$inRecord,[boolean]$includeTTL) {
		$trimmedData = if ($inRecord.data.length -gt 20) { $inRecord.data.Substring(0,17)+'...'} else {$inRecord.data}
		$ttlBlob = if ($includeTTL) { " $($inRecord.ttl) "} else { "" }
		return ('{0}{1}{2} {3}' -f $inRecord.name,$ttlBlob,$inRecord.type,$trimmedData)
	}
	static [string]getRecordAsString ([hashtable]$inRecord) {
		return [htmlUtils]::getRecordAsString($inRecord, $false)
	}

	static [string]getHumanReadableList ([string[]]$inList){
		if ($inList.Count -eq 1){
			return $inList[0]
		} elseif ($inList.Count -eq 0){
			return ''
		} elseif ($inList.Count -eq 2){
			return ($inList[0] + ' and ' +$inList[1])
		} else {
			return (($inList[0..($inList.Count-2)] -join ', ') + ' and ' + $inList[-1])
		}
	}

	static [string]getArrayAsHTMListItems([String[]]$list){
		$wrappedItems = @()
		foreach ($item in $list){
			$wrappedItems += ('<li>{0}</li>' -f $item)
		}
		return $wrappedItems -join ''
	}
}

class adUtils{

	static [DirectoryServices.DirectorySearcher]getADSISearcher([string]$query,[string[]]$propsToLoad){
		$searchRoot = (New-Object System.DirectoryServices.DirectoryEntry('LDAP://RootDSE')).rootDomainNamingContext
		if ($searchRoot){
			$searcher = [DirectoryServices.DirectorySearcher] $query
			$searcher.SearchRoot = 'GC://'+$searchRoot.toString()
			if ($propsToLoad.Count){
				$searcher.PropertiesToLoad.AddRange($propsToLoad) > $null
			}
			return $searcher
		} else {
			log ('ERROR: Could not query Active Directory: Could not initialize search root')
			return $null
		}
	}

	static [hashtable]queryAD([string]$adsiQuery,[string[]]$propsToLoad=@()){
		try {
			$propsToLoad += 'distinguishedname'
			$propsToLoad += 'samAccountName'
			$searcher = [adUtils]::getADSISearcher($adsiQuery, $propsToLoad)
			if ($searcher){
				$searchRes = $searcher.FindOne()
				if ($searchRes){
					$outMap = @{}
					foreach ($key in $searchRes.Properties.Keys){
						$prop = $searchRes.Properties.$key
						#most of the time it's just a single valued string. But for other cases, let's store as array
						$outMap[$key] = if ($prop.Count -eq 1) { $prop.item(0).toString() } else {@($prop)}
					}
					$outMap.dn = $outMap.distinguishedname # move it here, for ease of use
					return $outMap
				} else {
					logv2 "Found no results for ADSI query: $adsiQuery"
				}
			}
		} catch {
			log ("ERROR: Got error while querying Active Directory: $($_.Exception.Message)")
		}
		return $null #return null on errors, but empty map when nothing found
	}

	static [hashtable[]]findAll([string]$query, [string[]]$propsToLoad){
		try {
			$propsToLoad += 'distinguishedname'
			$propsToLoad += 'samAccountName'
			$searcher = [adUtils]::getADSISearcher($query, $propsToLoad)
			if ($searcher){
				$searchRes = $searcher.FindAll()
				if ($searchRes -and $searchRes.Count){
					$out = New-Object System.Collections.ArrayList
					foreach ($item in $searchRes){
						$outMap = @{}
						foreach ($key in $item.Properties.Keys){
							$prop = $item.Properties.$key
							#most of the time it's just a single valued string. But for other cases, let's store as array
							$outMap[$key] = if ($prop.Count -eq 1) { $prop.item(0).toString() } else {$prop}
						}
						$outMap.dn = $outMap.distinguishedname # move it here, for ease of use
						$out.Add($outMap)
					}
					return @($out)
				} else {
					logv2 "Found no results for ADSI query: $query"
					return @()
				}
			}
		} catch {
			log ("ERROR: Got error while querying Active Directory: $($_.Exception.Message)")
		}
		return @()
	}

	static [string]getUserSID ([string]$inUserName){
		$userParts = [mm]::parseUserName($inUserName)

		if (!$userParts){
			log "Username $inUserName incorrectly formatted"
			return $null
		}

		try{
			$account = new-object System.Security.Principal.NTAccount($userParts.domain,$userParts.name)
			$sid = $account.Translate([System.Security.Principal.SecurityIdentifier]).Value
			return $sid
		} catch [System.Security.Principal.IdentityNotMappedException] {
			#user not found in AD
			log "Could not find user $inUserName in AD"
			return $null
		}
	}

	static [hashtable]getUserByFilter([string]$filter, [string[]]$props){
		$props += 'msDS-PrincipalName'
		$props += 'mail'
		$result = [adUtils]::queryAD("(&($filter)(objectcategory=person)(objectclass=user))", $props)
		if ($result){
			$result.name = (?? $result.'msDS-PrincipalName' '').toLower()
			$result.Remove('msDS-PrincipalName')
			logv3 ('Found user {0}: {1}' -f $result.name, [mmJSON]::stringify($result))
			return $result
		} else {
			log 'ERROR: Could not retrieve user from AD: Could not initialize search root'
			return $null
		}
	}

	static [hashtable]getUserByName([string]$inName, [string[]]$propsToLoad){
		if ([mm]::isEmpty($inName)){
			logv2 "unable to get user with no name"
			return $null
		}
		$userParts = [mm]::parseUserName($inName)
		$sid = [adUtils]::getUserSID($inName)

		if (!$userParts){
			log "Username is incorrectly formatted. Needs to be in either of domain\user or user@domain formats."
		} elseif (!$sid){
			log "Could not find user $inName in AD (sid)"
		} else {
			return [adUtils]::getUserByFilter("objectsid=$sid", $propsToLoad)
		}
		return $null
	}

	static [hashtable]getUserByName([string]$inName){
		return [adUtils]::getUserByName($inName, 'mail')
	}

	static [hashtable]getUserByDN([string]$DN){
		return [adUtils]::getUserByFilter("distinguishedName=$DN", @())
	}

	static [hashtable]getUserByDN([string]$DN, [string[]]$props){
		return [adUtils]::getUserByFilter("distinguishedName=$DN", $props)
	}

	static [boolean]verifyUserByName([string]$inName){
		$userInfo = [adUtils]::getUserByName($inName,@('UserAccountControl'))
		if (!$userInfo){
			return $false
		}

		# verify the account is not disabled, even if found. If it's disabled, it's no longer valid
		$UACDisabledFlag = 0x02
		$isDisabled = ($userInfo.useraccountcontrol -band $UACDisabledFlag) -as [boolean]
		return !$isDisabled
	}

	static [hashtable]getGroupBySamAccountName ([string]$inGroupName, [string[]]$propsToLoad){
		#this queries by samAccountName because that's what M&M does (M&M doesn't find the group by cn/name, if it's different from samAccountName)
		return [adUtils]::queryAD("(&(objectCategory=group)(sAMAccountName=$inGroupName))", $propsToLoad)
	}

	static [hashtable]getGroup ([string]$inGroup, [string[]]$props){
		if ($inGroup.Contains('@')){
			# assume the user is querying by mail property
			return [adUtils]::getGroupByMail($inGroup,$props)
		}
		if ($inGroup.split('\').Count -eq 2){
			# assume it's domain\samAccountName format
			return [adUtils]::getGroupBySamAccountName($inGroup.split('\')[1], $props)
		}

		# match against either name (display name) or samAccountName
		return [adUtils]::queryAD("(&(objectCategory=group)(|(sAMAccountName=$inGroup)(name=$inGroup)(displayName=$inGroup)))", $props)
	}
	static [hashtable]getGroup([string]$inGroup){
		return [adUtils]::getGroup($inGroup, @('mail','name','cn'))
	}

	static [boolean]verifyGroup([string]$inGroup){
		return $null -ne [adUtils]::getGroup($inGroup)
	}

	static [hashtable]getGroupByMail ([string]$inPDL, [string[]]$propsToLoad){
		$groupEmail = $inPDL
		if (!$groupEmail.Contains('@')){
			# use domain in email address, else from setting, else use user dns domain
			$defaultEmailDomain = ?? $Global:mmConfig.AD.DNSDomain $ENV:USERDNSDOMAIN
			$groupEmail += ("@$defaultEmailDomain")
		}
		$propsToLoad += 'mail'
		return [adUtils]::queryAD("(&(objectCategory=group)(mail=$groupEmail))", $propsToLoad)
	}

	static [hashtable]getGroupByMail([string]$inPDL){
		return [adUtils]::getGroupByMail($inPDL, @('name', 'mail'))
	}

	static [boolean]verifyADGroupByMail([string]$inPDL){
		return $null -ne [adUtils]::getADGroupByMail($inPDL)
	}

	static [boolean]isUserMemberInGroupByDN([string]$inUserName, [string]$inGroupDN){
		$userInfo = [adUtils]::getUserByName($inUserName)
		if (!$userInfo){
			log ("User $inUserName could not be found in AD")
			return $false
		}

		$searcher = New-Object adsisearcher "(memberof:1.2.840.113556.1.4.1941:=$inGroupDN)"
		$searcher.SearchRoot = "GC://{0}" -f $userInfo.dn
		$searcher.SearchScope = 'Base'
		$isMember =  $null -ne $searcher.FindOne()
		$msg = if ($isMember) {'Is a Member!'} else {'NOT a member!'}
		log $msg

		return $isMember
	}

	static [boolean]isUserMemberInGroup([string]$inUserName, [string]$inGroup){
		$groupInfo = [adUtils]::getGroup($inGroup)
		if (!$groupInfo){
			log ("Group $inGroup could not be found in AD")
			return $false
		}
		return [adUtils]::isUserMemberInGroupByDN($inUserName, $groupInfo.dn)
	}

	static [string[]]getAllGroupsForUser([string]$inUserName, [boolean]$recursive){
		$userInfo = [adUtils]::getUserByName($inUserName)
		if (!$userInfo){
			return @()
		}

		$recurseToken = if ($recursive) {":1.2.840.113556.1.4.1941:"} else {""}

		tic ("Retrieving all groups for $inUserName $recurseToken")
		$userDN = $userInfo.dn
		$DNSDomain = ?? $Global:mmConfig.AD.DNSDomain $ENV:USERDNSDOMAIN
		$queryStr= "(&(objectCategory=group)(mail=*@$DNSDomain)(member$recurseToken=$userDN))"
		$groupsMap = @{}
		$groups = @()
		try {
			$searcher = [adUtils]::getADSISearcher($queryStr, @('mail','cn'))
			if ($searcher) {
				$results = $searcher.FindAll()
				log ("got $($results.count) results")
				foreach ($result in $results) {
					if ([mm]::notEmpty($result.Properties.mail)) {
						$mail = [string]$result.Properties.mail
						if (!$groupsMap.ContainsKey($mail)){
							$entry = @{
								mail = $mail
								name = [string]$result.Properties.cn
								pdl  = [string]$result.Properties.mail
							}
							$groupsMap[$mail] = $entry
							$groups += $entry
						}
					}
				}
			}
		} catch [System.Security.Principal.IdentityNotMappedException] {
			throw "Unable to retrieve groups for user. Identity Not Mapped"
		}
		toc "Retrieved $($groups.Count) groups"
		return $groups
	}

	static [string[]]getAllGroupsForUser([string]$inUserName){
		return [adUtils]::getAllGroupsForUser($inUserName, $false)
	}

	static [string[]]getMutualGroupsForUsers([string[]]$inUserNames) {
		# first, fetch all groups for all users
		$groupsByUser = @{}
		foreach ($user in $inUserNames){
			$groupsByUser[$user] = [adUtils]::getAllGroupsForUser($user)
		}

		# then, using the groups for the first user, check which of them are in all of the group lists - those are mutual
		# we could optimize this by eliminitating non mutual groups after fetching groups for each user, breaking once the list has become empty (fetching more groups for users wont help)
		# but we'll leave that for a future improvement
		$firstUserGroups = $groupsByUser[$inUserNames[0]]
		$mutualGroups = @()
		foreach ($group in $firstUserGroups){
			$isMutual = $true
			foreach ($user in $inUserNames){
				$inGroups = $false
				foreach ($grp in $groupsByUser[$user]){
					if ($grp.pdl -eq $group.pdl){
						$inGroups = $true
					}
				}
				if (!$inGroups){
					$isMutual = $false
					break
				}
			}
			if ($isMutual){
				$mutualGroups += $group
			}
		}
		return $mutualGroups
	}

	static [hashtable[]]getUsersInGroupByDN([string]$groupDN, [string[]]$propsToLoad, [boolean]$recursive){
		if ($Global:mmCache.hasCachedGroupMembersForDN($groupDN)){
			$users = $Global:mmCache.getGroupMembersForDN($groupDN)
			logv2 "Using cached members for group, found $($users.Count)"
			return $users
		}
		$propsToLoad += 'msDS-PrincipalName'
		$users = [adUtils]::findAll("(&(memberof=$groupDN)(objectclass=user))",$propsToLoad)

		if (!$recursive){
			return $users
		}

		# now add members of all groups, that are a member of this group
		# that is, a single level "recursion"
		$usersByDN = $users | mapBy 'dn'
		$groups = [adUtils]::findAll("(&(memberof=$groupDN)(objectclass=group))",'member')
		$maxGroupSize = ?? $Global:mmConfig.AD.maxGroupSize 100
		foreach($group in $groups){
			if ($group.member.count -and $group.member.count -lt $maxGroupSize) { # empty or very large groups ignored
				# NOTE: member is not populated for Primary groups or Domain Users groups. So we don't need to worry about those here
				$groupUsers = [adUtils]::findAll("(&(memberof=$($group.dn))(objectclass=user))",$propsToLoad)
				foreach ($groupUser in $groupUsers){
					if (!$usersByDN.ContainsKey($groupUser.dn)){
						$users += $groupUser
						$usersByDN[$groupUser.dn] = $groupUser
					}
				}

			}
		}
		$Global:mmCache.cacheGroupMembers($groupDN,$users)
		return $users
	}
	static [hashtable[]]getUsersInGroupByDN([string]$groupDN, [string[]]$propsToLoad){
		return [adUtils]::getUsersInGroupByDN($groupDN, $propsToLoad, $false)
	}
	static [hashtable[]]getUsersInGroupByDN([string]$groupDN){
		return [adUtils]::getUsersInGroupByDN($groupDN, @())
	}
	static [hashtable[]]getUsersInGroup([string]$group, [string[]]$propsToLoad, [boolean]$recursive){
		$adGroup = [adUtils]::getGroup($group)
		return [adUtils]::getUsersInGroupByDN($adGroup.dn, $propsToLoad, $recursive)
	}
	static [hashtable[]]getUsersInGroup([string]$group, [string[]]$propsToLoad){
		return [adUtils]::getUsersInGroup($group, $propsToLoad, $false)
	}
	static [hashtable[]]getUsersInGroup([string]$group){
		return [adUtils]::getUsersInGroup($group, @())
	}

	static [string]getEmailForUser([string]$username){
		logv2 "Getting email for user $username"
		return [adUtils]::getUserByName($username, @('mail')).mail
	}
	static [string[]]getEmailsForUsers([string[]]$usernames){
		$emails = @()
		foreach ($user in $usernames){
			$mail = [adUtils]::getEmailForUser($user)
			if ([mm]::notEmpty($mail)){
				$emails += $mail
			} else {
				log "$user not found or mail attribute not set"
			}
		}
		return $emails
	}

	static [hashtable]getManagerForUser([string]$username){
		# retrieve the user from AD
		$userInfo = [adUtils]::getUserByName($username, @('manager'))
		if (!$userInfo){
			log ('User {0} not found in AD. Assumed to be an internal M&M user' -f $username)
			return $null
		}

		if (!$userInfo.manager){
			log ('User {0} found in AD as {1}, but has no manager registered' -f $username, $userInfo.name)
			return $null
		}

		log ('Found {0} as "Manager" for {1}' -f $userInfo.manager, $username)
		$managerInfo = [adUtils]::getUserByDN($userInfo.manager)
		if (!$managerInfo -or !$managerInfo.name){
			log ('Unable to look up info for manager ' + $userInfo.manager)
			return $null
		}
		return $managerInfo
	}

}

Enum LogSeverity {
	normal = 1
	warning = 2
	debug = 3
	error = 4
}

class Logger {
	[string]$logfilename
	[boolean]$logToConsole
	[int]$verbosity

	[datetime]$lastTimer
	[string]$lastTimerAction
	[string]$lastMsg

	Logger([string]$logfile, [boolean]$logToConsole, [int]$verbosity){
		if ($logfile.Contains('/')) {
			# Seems to be path and not a filename. Assume callee knows what he's doing and just use this path as is
			$this.logfilename = $logfile
		} else {
			# logs are logged to ./logs/ relative to utils.ps1
			$logsPath = Join-Path $PSScriptRoot 'logs'
			if (!(Test-Path $logsPath)){
				mkdir $logsPath
			}
			$this.logfilename = Join-Path $logsPath $logfile
		}
		$this.logToConsole = $logToConsole
		$this.verbosity = $verbosity
	}

	static logToFile([string]$msg, [string]$theLogFile, [boolean]$logToConsole, [LogSeverity]$severity) {
		$timeStr = [mm]::getCurrentDate()

		if (!$msg) {
			$outMsg = ' '
		} else {
			$outMsg = "{0} ({2}): {1}" -f $timeStr,$msg,$Global:PID
		}

		if ($logToConsole) {
			if ($severity -eq [LogSeverity]::normal){
				Write-Host $outMsg
			} elseif ($severity -eq [LogSeverity]::warning){
				Write-Host -ForegroundColor Yellow -BackgroundColor Black $outMsg
			} elseif ($severity -eq [LogSeverity]::debug){
				Write-Host -ForegroundColor Cyan -BackgroundColor Black $outMsg
			} elseif ($severity -eq [LogSeverity]::error){
				Write-Host -ForegroundColor Red -BackgroundColor Black $outMsg
			}
		}

		if ([mm]::notEmpty($theLogFile)){
			$retry = 0
			while ($retry++ -lt 200) {
				try {
					$outMsg | Out-File -Append $theLogFile
					break
				} catch {
					start-sleep -m 25
				}
			}
		}
	}
	static logToFile([string]$msg, [string]$theLogFile, [boolean]$logToConsole){
		[Logger]::logToFile($msg, $theLogFile, $logToConsole, [LogSeverity]::normal)
	}

	log([string]$msg,[LogSeverity]$severity){
		if ($msg -ne $this.lastMsg){ # don't repeat log msgs
			[Logger]::logToFile($msg, $this.logfilename, $this.logToConsole, $severity)
			$this.lastMsg = $msg
		}
	}

	log([string]$msg){
		$this.log($msg, [LogSeverity]::normal)
	}

	logv2([string]$msg){
		if ($this.verbosity -gt 1){
			$this.log($msg, [LogSeverity]::warning)
		}
	}

	logv3([string]$msg){
		if ($this.verbosity -gt 2){
			$this.log($msg, [LogSeverity]::debug)
		}
	}

	logException([string]$msg){
		$location =  '{0}:{1}' -f $_.InvocationInfo.PSCommandPath.split('\')[-1], $_.InvocationInfo.ScriptLineNumber
		$msg = '** Exception in {0} - {1}: {2}' -f $location,$msg,$_.Exception.GetBaseException().Message
		$this.log($msg, [LogSeverity]::error)
	}

	tic([string]$action){
		$this.log($action)
		$this.lastTimer = Get-Date
		$this.lastTimerAction = $action
	}

	toc([string]$customAction){
		$duration = (get-date)-$this.lastTimer
		if ($customAction){
			$this.log(('{0} in {1} msec' -f $customAction, [int]$duration.TotalMilliseconds))
		} else {
			$this.log(('{0} took {1} msec' -f $this.lastTimerAction, [int]$duration.TotalMilliseconds))
		}
		$this.lastTimer = Get-Date
	}
}

class Cache {
	[hashtable]$groupMembersByDN

	Cache(){
		$this.groupMembersByDN = @{}
	}

	[void]cacheGroupMembers([string]$groupDN, [hashtable[]]$members){
		$this.groupMembersByDN[$groupDN]=$members
	}
	[boolean]hasCachedGroupMembersForDN([string]$groupDN){
		return $this.groupMembersByDN.ContainsKey($groupDN)
	}
	[hashtable[]]getGroupMembersForDN([string]$groupDN){
		return $this.groupMembersByDN[$groupDN]
	}
}

